/*  Torpatch - A program to construct 'tangles' from torus sections.
               For use with the Persistence of Vision ray tracer.
    Copyright (C) 1995 Mark Mackey.
    C version Copyright (C) 1996 Mark Mackey and Nathan Beddes.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    You can contact the original author, Mark Mackey,  at 
    mdm1004@cus.cam.ac.uk, and the translator, Nathan Beddes, at
    Nathan.Beddes@m.cc.utah.edu.
*/


/* WARNING! This code is a C translation of the original Turbo Pascal code.
   The original code was fairly 'orrible, but the C translation is worse.
   This code is largely uncommented, unreadable, blech and yeaaaaggghh, and I 
   hereby disclaim all responsibility for it even though I wrote it :).

   It _should_, however, compile fairly cleanly straight out of the box.
*/
   
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>

typedef double real;

typedef struct vector {
  real x, y, z;
} vector;

typedef vector point;
/*Not strictly necessary, but the distinction is sometimes useful*/

typedef struct pointlisttype {
  vector pt;
  struct pointlisttype *next;
} pointlisttype;

static long N = 150;
static vector a1, a2, a3, intangent, outtangent, centre, min, max;
static real radius, xang, yang, max_radius;
static FILE *infile, *outfile;
static unsigned char showpoints, showtangents, tangentmode, intersection;
static int pnum;
static char s[256];
static char name[41];
static char new[2];
static char ch;
static long num_random, i, repeats;
static pointlisttype pointlisthead, *plist;


int eoln(f)       /*Thanks to Dave Gillespie for this function*/
FILE *f;	  /*as well as for the great p2c program!     */
{
    int ch;

    ch = getc(f);
    if (ch == EOF)
        return 1;
    ungetc(ch, f);
    return (ch == '\n');
}


static unsigned char exist(filename)
char *filename;
{
  /*Returns true if file "filename" exists in current directory*/
  unsigned char Result;
  FILE *fil;
  if (fopen(filename, "r+") == NULL)
	Result = 0;	
  else  
	Result = 1;
  return Result;
}

static void cross(v1, v2, result)
vector v1, v2, *result;
{
  /*Vector cross product*/
  result->x = v1.y * v2.z - v2.y * v1.z;
  result->y = v2.x * v1.z - v1.x * v2.z;
  result->z = v1.x * v2.y - v2.x * v1.y;
}


static double dot(v1, v2)
vector v1, v2;
{
  /*Vector dot product*/
  return (v1.x * v2.x + v1.y * v2.y + v1.z * v2.z);
}


static void normalise(v)
vector *v;
{
  /*Normalises a vector*/
  real r;

  r = sqrt(v->x * v->x + v->y * v->y + v->z * v->z);
  if (r == 0)
    return;
  v->x /= r;
  v->y /= r;
  v->z /= r;
}


static void rotatex(v, ang, result)
vector v;
double ang;
vector *result;
{
  /*rotate a vector/point around x axis by ang radians*/
  real sinang, cosang;

  sinang = sin(ang);
  cosang = cos(ang);
  result->x = v.x;
  result->y = v.y * cosang + v.z * sinang;
  result->z = v.z * cosang - v.y * sinang;
}


static void rotatey(v, ang, result)
vector v;
double ang;
vector *result;
{
  /*rotate a vector/point around y axis by ang radians*/
  real sinang, cosang;

  sinang = sin(ang);
  cosang = cos(ang);
  result->x = v.x * cosang + v.z * sinang;
  result->y = v.y;
  result->z = v.z * cosang - v.x * sinang;
}


static void addvectors(v1, v2, result)
vector v1, v2, *result;
{
  /*Add two vectors*/
  result->x = v1.x + v2.x;
  result->y = v1.y + v2.y;
  result->z = v1.z + v2.z;
}


static void subtractvectors(v1, v2, result)
vector v1, v2, *result;
{
  /*Subtract two vectors*/
  result->x = v1.x - v2.x;
  result->y = v1.y - v2.y;
  result->z = v1.z - v2.z;
}


static void scalevector(v, scalefactor)
vector *v;
double scalefactor;
{
  /*Vector by scalar multiplication*/
  v->x *= scalefactor;
  v->y *= scalefactor;
  v->z *= scalefactor;
}


static void shiftvector(v, shift)
vector *v;
double shift;
{
  /*scalar addition to vector*/
  v->x += shift;
  v->y += shift;
  v->z += shift;
}


static double distance(p1, p2)
vector p1, p2;
{
  /*Returns the distance between two points in 3-space*/
  double TEMP, TEMP1, TEMP2;

  TEMP = p1.x - p2.x;
  TEMP1 = p1.y - p2.y;
  TEMP2 = p1.z - p2.z;
  return sqrt(TEMP * TEMP + TEMP1 * TEMP1 + TEMP2 * TEMP2);
}


static void writevector(heading, v)
char *heading;
vector v;
{
  /*Writes a vector to stdout*/
  printf("%s x:%12.6f y:%12.6f z:%12.6f\n", heading, v.x, v.y, v.z);
}


static void Find2DCentre(q1, q2, q3, result)
vector q1, q2, q3, *result;
{
  /*Finds the circumcentre of q1..3 (in two dimensions). Assumes that
    q1..q3 have the same z-coordinate.*/
  vector r1, r2;
  real m1, m2;

  addvectors(q1, q2, &r1);
  scalevector(&r1, 0.5);
  addvectors(q2, q3, &r2);
  scalevector(&r2, 0.5);
  /*r1 and r2 are the midpoints of the line segments
                          q1->q2 and q2->q3*/

  if (q1.x - q2.x != 0)
    m1 = (q1.y - q2.y) / (q1.x - q2.x);   /*gradient of first line segment*/
  else
    m1 = q1.y - q2.y;
  if (q2.x - q3.x != 0)
    m2 = (q2.y - q3.y) / (q2.x - q3.x);   /*gradient of second line segment*/
  else
    m2 = q2.y - q3.y;
  if (m1 == 0)
    m1 = 1.00e-10;
  if (m2 == 0)
    m2 = 1.01e-10;
  if (m1 == m2)   /*should really output cylinder, but..*/
    m1 *= 1.000000001;

  result->x = (r2.x / m2 - r1.x / m1 + r2.y - r1.y) / (1 / m2 - 1 / m1);
  /*basic Cartesian geometry :)*/
  result->y = r1.y - 1 / m1 * (result->x - r1.x);
  result->z = q1.z;   /*q1..3 lie in same z plane, remember?*/
}


static void Find2DCentre2(q1, q2, tangent, result)
vector q1, q2, tangent, *result;
{
  /*Finds the circumcentre of q1 and q2 given the tangent vector at q1
    in two dimensions. Assumes that the circumcircle lies in the xy plane
    ie that q1 and q2 have the same z coordinate and that the tangent
    vector has zero z component*/
  vector r1, r2;
  real m1, m2;

  r1 = q1;
  addvectors(q1, q2, &r2);
  scalevector(&r2, 0.5);   /*r1 and r2 are the endpoints of two radii*/
  if (tangent.x != 0)   /*close enough*/
    m1 = tangent.y / tangent.x;   /*gradient of tangent*/
  else
    m1 = tangent.y;
  if (q1.x - q2.x != 0)   /*close enough*/
    m2 = (q1.y - q2.y) / (q1.x - q2.x);   /*gradient of chord q1->q2*/
  else
    m2 = q1.y - q2.y;
  if (m1 == 0)   /*"*/
    m1 = 1.0e-10;
  if (m2 == 0)   /*"*/
    m2 = 1.0e-10;
  if (m1 == m2)   /*should really output cylinder, but..*/
    m1 *= 1.000000001;
  result->x = (r2.x / m2 - r1.x / m1 + r2.y - r1.y) / (1 / m2 - 1 / m1);
  /*more elementary geometry :) */
  result->y = r1.y - 1 / m1 * (result->x - r1.x);
  result->z = q1.z;   /*assume that q1..3 lie in same z plane*/
}


static void FindTorusCentre(p1, p2, p3, centre, intangent, outtangent, radius,
			    xang, yang)
vector p1, p2, p3, *centre, *intangent, *outtangent;
double *radius, *xang, *yang;
{

  /*takes 3 points in 3-space as input. Returns the centre and radius of
    a torus in the xy plane and two angles to rotate this torus so that
    it will pass through the original 3 points. Also returns the tangent
    vectors from the first and third points (untransformed ie in original
    coordinate frame)

    The inwards and outgoing tangents are guaranteed to point in opposite
    directions along the torus arc.*/
  vector r1, r2;   /*r1=p1->p2, r2=p2->p3*/
  vector n;   /*normal to plane of p1..3*/
  vector q1, q2, q3;
      /*p1..3 rotated so their plane is parallel to xy plane*/
  vector test;
  real m;
  unsigned char positive;

  /*records the relative direction of the inwards
                     and outwards tangents*/

  subtractvectors(p2, p1, &r1);   /*r1 = p1->p2*/
  subtractvectors(p3, p2, &r2);   /*r2 = p2->p3*/
  cross(r1, r2, &n);   /*normal vector to p1,p2,p3 plane*/
  if (n.z != 0)
    *xang = atan(-(n.y / n.z));
  else
    *xang = atan(-n.y);
  rotatex(n, *xang, &n);   /*rotate n to eliminate x component*/
  if (n.z != 0)
    *yang = atan(-(n.x / n.z));
  else
    *yang = atan(-n.x);
  rotatey(n, *yang, &n);   /*rotate n to eliminate y component*/
  rotatex(p1, *xang, &q1);
  rotatey(q1, *yang, &q1);
  rotatex(p2, *xang, &q2);
  rotatey(q2, *yang, &q2);
  rotatex(p3, *xang, &q3);   /*rotate p1..p3 by these angles to*/
  rotatey(q3, *yang, &q3);   /*obtain q1..q3 with constant z-coords*/
  Find2DCentre(q1, q2, q3, centre);   /*Find circumcentre in 2D*/
  *radius = distance(q1, *centre);   /*Radius of torus*/

  m = (centre->y - q1.y) / (centre->x - q1.x);
  /*gradient of 2D vector from centre to
                                       1st point*/
  intangent->x = 1.0;
  if (m != 0)
    intangent->y = -1 / m;
  else
    intangent->y = -1.0e10;
  intangent->z = 0.0;
  rotatey(*intangent, -*yang, intangent);
  rotatex(*intangent, -*xang, intangent);
  /*transform tangent vector to original
                                       coordinate frame*/

  subtractvectors(q1, *centre, &test);   /*test = centre -> q1*/
  cross(test, *intangent, &test);
  /*which direction is this tangent
                                       pointing?*/
  positive = (test.z > 0);
  /*for comparison with outtangent: are they
                         pointing in same/opp dir?*/

  m = (centre->y - q3.y) / (centre->x - q3.x);
  /*gradient of 2D vector from centre to
                                       3rd point*/
  outtangent->x = 1.0;
  outtangent->y = -1 / m;

  outtangent->z = 0.0;
  subtractvectors(q3, *centre, &test);   /*test = centre -> q3*/

  cross(test, *outtangent, &test);
  /*which direction is this tangent
                                       pointing?*/
  if (positive && test.z > 0 || !positive && test.z < 0)
    scalevector(outtangent, -1.0);
  /*make sure intangent and outtangent point in opp. dirs*/

  rotatey(*outtangent, -*yang, outtangent);
  rotatex(*outtangent, -*xang, outtangent);
  /*transform tangent vector to original coordinate frame*/
}


static void FindTorusCentre2(p1, p2, intangent, centre, outtangent, radius,
			     xang, yang)
vector p1, p2, intangent, *centre, *outtangent;
double *radius, *xang, *yang;
{
  /*takes 2 points in 3-space and an initial tangent vector (pointing away
    from the arc) as input. Returns the centre and radius of a torus in
    the xy plane and two angles to rotate this torus so that it will pass
    through the original 3 points. Also returns the tangent vector from
    the second point (untransformed ie in original coordinate frame).

    The inwards and outgoing tangents are guaranteed to point in opposite
    directions along the torus arc.*/
  vector r;   /*r1=p1->p2*/
  vector n;   /*normal to plane of p1->p2 and intangent vector*/
  vector q1, q2;   /*transformed p1,p2 into xy plane*/
  vector test;
  real m;
  unsigned char positive;

  /*records the relative direction of the inwards and outwards tangents*/


  subtractvectors(p2, p1, &r);   /*r = p1->p2*/
  cross(intangent, r, &n);   /*normal vector to plane of intangent and r*/
  if (n.z != 0)
    *xang = atan(-(n.y / n.z));
  else
    *xang = atan(-n.y);
  rotatex(n, *xang, &n);   /*rotate normal to eliminate x component*/
  if (n.z != 0)
    *yang = atan(-(n.x / n.z));
  else
    *yang = atan(-n.x);
  rotatey(n, *yang, &n);   /*rotate normal to eliminate y component*/
  rotatex(p1, *xang, &q1);
  rotatey(q1, *yang, &q1);
  rotatex(p2, *xang, &q2);   /*rotate p1 and p2 by these angles to*/
  rotatey(q2, *yang, &q2);   /*obtain q1 and q2 with constant z-coords*/
  rotatex(intangent, *xang, &intangent);
  rotatey(intangent, *yang, &intangent);
  /*rotate intangent into same coordinate frame*/

  Find2DCentre2(q1, q2, intangent, centre);
      /*find circumcentre in 2 dimensions*/
  *radius = distance(q1, *centre);   /*radius of circle*/

  subtractvectors(q1, *centre, &test);   /*test = centre -> q1*/
  cross(test, intangent, &test);
      /*what direction is this tangent pointing?*/
  positive = (test.z > 0);
  /*for comparison with outtangent: pointing in
                         same/opp dir?*/

  m = (centre->y - q2.y) / (centre->x - q2.x);
      /*gradient of radius to 2rd point*/
  outtangent->x = 1.0;
  outtangent->y = -1 / m;
  outtangent->z = 0.0;
  subtractvectors(q2, *centre, &test);   /*vector from centre to endpoint*/
  cross(test, *outtangent, &test);
      /*what direction is this tangent pointing?*/

  if (positive && test.z > 0 || !positive && test.z < 0)
    scalevector(outtangent, -1.0);
  /*make sure intangent and outtangent point in opp. dirns*/

  rotatey(*outtangent, -*yang, outtangent);
  rotatex(*outtangent, -*xang, outtangent);
      /*transform to original coord frame*/

}


static void Paramhelp()
{
  /*Print help on TORPATCH syntax*/
  printf(
    "\nSyntax is torpatch [-t] [-p] [-c] [-n name] <input file> <output file>\n\n");
  printf("The -t flag signifies that the first two points define a\n");
  printf("  tangent to the first torus. If this flag is not specified\n");
  printf("  then the first torus patch is defined to pass through the\n");
  printf("  first three points.\n\n");
  printf("The -p flag indicates that a structure called Points containing\n");
  printf("  the control points is to be included in the output.\n\n");
  printf("The -c flag indicates that tangent cylinders are to be\n");
  printf("  placed at each control point. The red cylinders point\n");
  printf("  in the 'forwards' direction. This option is mainly\n");
  printf("  included for debugging purposes.\n\n");
  printf("The -n flag allows you to specify the name of the torus patch. The\n");
  printf("  default name is Torus_Patch. Textures have '_Texture' appended.\n");
  printf("\nTorpatch version 1.0, Copyright (C) 1996 Mark Mackey\n");
  printf("Torpatch comes with ABSOLUTELY NO WARRANTY. This is free software,\n");
  printf("and you are welcome to redistribute it under certain conditions;\n");
  printf("see the file README.GNU for details.\n");

  exit(0);
}


static void OutputTangent(a1, intangent, a3, outtangent)
vector a1, intangent, a3, outtangent;
{
  /*Print tangent object*/
  normalise(&outtangent);
  normalise(&intangent);
  fprintf(outfile, "    object{\n");
  fprintf(outfile, "      cylinder{\n");
  fprintf(outfile, "        <%12.6f,%12.6f,%12.6f>\n", a1.x, a1.y, a1.z);
  fprintf(outfile, "        <%12.6f,%12.6f,%12.6f> +\n", a1.x, a1.y, a1.z);
  fprintf(outfile, "          (<%12.6f,%12.6f,%12.6f> * Tangent_Length)\n",
	  intangent.x, intangent.y, outtangent.z);
  fprintf(outfile, "      Tangent_Radius\n");
  fprintf(outfile, "      texture{Tangent_Texture_In}\n");
  fprintf(outfile, "      }\n");
  fprintf(outfile, "    }\n");
  fprintf(outfile, "    object{\n");
  fprintf(outfile, "      cylinder{\n");
  fprintf(outfile, "        <%12.6f,%12.6f,%12.6f>\n", a3.x, a3.y, a3.z);
  fprintf(outfile, "        <%12.6f,%12.6f,%12.6f> +\n", a3.x, a3.y, a3.z);
  fprintf(outfile, "          (<%12.6f,%12.6f,%12.6f> * Tangent_Length)\n",
	  outtangent.x, outtangent.y, outtangent.z);
  fprintf(outfile, "        Tangent_Radius\n");
  fprintf(outfile, "        texture{Tangent_Texture_Out}\n");
  fprintf(outfile, "      }\n");
  fprintf(outfile, "    }\n");
}


static char *snip(Result, s)
char *Result;
char *s;
{
  char STR1[256];

  while (*s != '\0' && (s[0] == ' ' || s[0] == ','))
    strcpy(&s[0], &s[1]);
  i = 0;
  do {
    i++;
  } while (i <= strlen(s) && s[i - 1] != ' ' && s[i - 1] != ',');
  sprintf(Result, "%.*s", (int)(i - 1), s);
  strcpy(&s[0], &s[i]);
  return Result;
}


static unsigned char readcoords(f, x, y, z, s)
FILE **f;
double *x, *y, *z;
char *s;
{
  /*returns success value*/
  unsigned char Result;
  char s2[256], s3[256];
  long i;
  char *TEMP;
  
  /*Here be kludges...*/

  int ch;

  ch = fgetc(*f);
  while ((ch == '\n') && (ch != EOF))
    ch = fgetc(*f);
  ungetc(ch, *f);

  /*End of kludge*/

  Result = 1;
  fgets(s2, 256, *f);
  TEMP = (char *) strchr(s2, '\n');
  if (TEMP != NULL)
    *TEMP = 0;
  strcpy(s, s2);
  snip(s3, s);
  i = (sscanf(s3, "%lg", x) == 0);
  if (i != 0)
    Result = 0;
  snip(s3, s);
  i = (sscanf(s3, "%lg", y) == 0);
  if (i != 0)
    Result = 0;
  snip(s3, s);
  i = (sscanf(s3, "%lg", z) == 0);
  if (i != 0 || *s3 == '\0')
    Result = 0;
  strcpy(s, s2);
  return Result;
}


static unsigned char GetRandomPoint(a2, min, max, a1, limit)
vector *a2, min, max, a1;
double limit;
{
  /*returns a point within limit of a1, inside box <min, max> if possible*/
  unsigned char Result;
  vector temp;

  a2->x = drand48() * (max.x - min.x) + min.x;
  a2->y = drand48() * (max.y - min.y) + min.y;
  a2->z = drand48() * (max.z - min.z) + min.z;
  Result = 0;
  temp = min;
  shiftvector(&temp, -limit);
  if (temp.x >= a1.x || temp.y >= a1.y || temp.z >= a1.z)
    return Result;
  temp = max;
  shiftvector(&temp, limit);
  if (temp.x <= a1.x || temp.y <= a1.y || temp.z <= a1.z)
    return Result;
  i = 0;
  do {
    i++;
    a2->x = drand48() * (max.x - min.x) + min.x;
    a2->y = drand48() * (max.y - min.y) + min.y;
    a2->z = drand48() * (max.z - min.z) + min.z;
  } while (!((i > 10000) | (distance(*a2, a1) <= limit)));
  if (i < 10000)
    return 1;
  return Result;
}


static void addtopointlist(p)
vector p;
{
  plist->pt = p;
  plist->next = (pointlisttype *)malloc(sizeof(pointlisttype));
  plist->next->next = NULL;
  plist = plist->next;
}


static void OutputTorus(a1, a2, centre, radius, xang, yang, intangent,
			intersection)
vector a1, a2, centre;
double radius, xang, yang;
vector intangent;
unsigned char intersection;
{
  /*Prints one torus patch object from two endpoints, centre, rotation angles
    and ingoing tangent. The intersection parameter controls the clipping planes
    and depends on whether the arc of the torus used subtends more than
    180 degrees*/
  fprintf(outfile, "    object{\n");
  fprintf(outfile, "      torus{ %12.6f, Minor_Radius}\n", radius);
  fprintf(outfile, "      rotate x*90\n");
  fprintf(outfile, "      bounded_by{\n");
  fprintf(outfile, "        box{\n");
  fprintf(outfile,
    "          <%6.3f-Minor_Radius, %6.3f-Minor_Radius, -Minor_Radius*1.001>\n",
    -radius * 1.001, -radius * 1.001);
  fprintf(outfile,
    "          <%6.3f+Minor_Radius, %6.3f+Minor_Radius, Minor_Radius*1.001>\n",
    radius * 1.001, radius * 1.001);
  fprintf(outfile, "        }\n");
  fprintf(outfile, "      }\n");
  fprintf(outfile, "      translate< %12.6f,%12.6f,%12.6f>\n",
	  centre.x, centre.y, centre.z);
  fprintf(outfile, "      rotate y*(%12.6f)\n", -yang * 180 / M_PI);
  fprintf(outfile, "      rotate x*(%12.6f)\n", xang * 180 / M_PI);

  /*Note: the change in sign in xang is because I am using a right-handed
     coordinate system within this program, but the output is for the
     POV default left-handed coordinate system. The rotation in the
     x direction thus has to be reversed.*/

  fprintf(outfile, "      texture{%s_Texture}\n", name);
  fprintf(outfile, "      clipped_by{\n");
  if (intersection)
    fprintf(outfile, "        intersection{\n");
  else
    fprintf(outfile, "        union{\n");
  fprintf(outfile, "          plane{ <%12.6f,%12.6f,%12.6f> 0\n",
	  intangent.x, intangent.y, intangent.z);
  fprintf(outfile, "            translate <%12.6f,%12.6f,%12.6f>}\n",
	  a1.x, a1.y, a1.z);
  fprintf(outfile, "          plane{ <%12.6f,%12.6f,%12.6f> 0\n",
	  outtangent.x, outtangent.y, outtangent.z);
  fprintf(outfile, "            translate <%12.6f,%12.6f,%12.6f>}\n",
	  a2.x, a2.y, a2.z);
  fprintf(outfile, "        }\n");
  fprintf(outfile, "      }\n");
  fprintf(outfile, "    }\n");
}


main(argc, argv)
int argc;
char *argv[];
{
  long randseed;
  char STR1[256];
  pointlisttype *WITH;
  outfile = NULL;
  infile = NULL;
  randseed = 2;
  if (argc < 3)
    Paramhelp();
  tangentmode = 0;
  showtangents = 0;
  showpoints = 0;
  strcpy(name, "Torus_Patch");
  pnum = 1;
  plist = &pointlisthead;
  num_random = 10000;
  strcpy(new, "\n");
  strcpy(s, argv[pnum]);
  while (s[0] == '-' || s[0] == '/') {   /*parse parameters*/
    switch (toupper(s[1])) {

    case 'T':
      tangentmode = 1;
      break;

    case 'C':
      showtangents = 1;
      break;

    case 'P':
      showpoints = 1;
      break;

    case 'N':
      strcpy(name, argv[pnum + 1]);
      pnum++;
      break;

    default:
      Paramhelp();
      break;
    }
    pnum++;
    strcpy(s, argv[pnum]);
  }
  if (pnum>=(argc-1))
    Paramhelp();
  strcpy(STR1,argv[pnum]);
  infile = fopen(STR1, "r+");
  rewind(infile);
  if (fopen(argv[pnum], "r+") == NULL) {
    printf("File %s not found.\n", argv[pnum]);
    Paramhelp();
  }
  strcpy(STR1, argv[pnum + 1]);
  if (exist(STR1)) {
    printf("File %s exists. Overwrite (y/n)?\n", argv[pnum + 1]);
    do {
      ch = toupper(getchar());
    } while (ch != 'Y' && ch != 'N');
    if (ch == 'N')
      exit(0);
  }
  outfile = fopen(STR1, "w");
  if (outfile != NULL)
    rewind(outfile);
  else
    outfile = tmpfile();
  printf("Joining points:\n");
  do {
  } while (!(readcoords(&infile, &a1.x, &a1.y, &a1.z, s) || (feof(infile) != 0)));
  do {
  } while (!(readcoords(&infile, &a2.x, &a2.y, &a2.z, s) || (feof(infile) != 0)));
  do {
  } while (!(readcoords(&infile, &a3.x, &a3.y, &a3.z, s) || (feof(infile) != 0)));
  if (tangentmode) {
    addtopointlist(a2);
    addtopointlist(a3);
    writevector("Initial tangent control point: ", a1);
    writevector("Initial tangent at: ", a2);
  } else {
    addtopointlist(a1);
    addtopointlist(a2);
    addtopointlist(a3);
    writevector("Initial control point", a1);
    writevector("Initial control point", a2);
  }
  writevector("Initial control point", a3);
  if (!tangentmode)
    FindTorusCentre(a1, a2, a3, &centre, &intangent, &outtangent, &radius,
		    &xang, &yang);
  else {
    subtractvectors(a1, a2, &intangent);
    FindTorusCentre2(a2, a3, intangent, &centre, &outtangent, &radius, &xang,
		     &yang);
  }
  fprintf(outfile,
      "/*-----------------------------------------------------------------------\\\n");
  fprintf(outfile,
      "|  This file was generated by TORPATCH v1.0 copyright 1995 Mark Mackey   |\n");
  fprintf(outfile,
      "|                 C version by Nathan Beddes, 1996.                      |\n");
  fprintf(outfile,
      "\\-----------------------------------------------------------------------*/\n");
  putc('\n', outfile);
  fprintf(outfile, "#declare Minor_Radius=0.05\n");
  fprintf(outfile, "#declare %s_Texture=texture{pigment{Green}}\n",
		 name);
  putc('\n', outfile);
  if (showpoints) {
    fprintf(outfile, "#declare Point_Radius=0.1\n");
    fprintf(outfile,
		   "#declare Point_Texture=texture{pigment{White}}\n");
    putc('\n', outfile);
  }
  if (showtangents) {
    fprintf(outfile, "#declare Tangent_Radius=0.05\n");
    fprintf(outfile, "#declare Tangent_Length=1\n");
    fprintf(outfile, "#declare Tangent_Texture_In=\n");
    fprintf(outfile, "  texture{\n");
    fprintf(outfile, "    pigment{Red}\n");
    fprintf(outfile, "  }\n");
    fprintf(outfile, "#declare Tangent_Texture_Out=\n");
    fprintf(outfile, "  texture{\n");
    fprintf(outfile, "    pigment{Blue}\n");
    fprintf(outfile, "  }\n");
    putc('\n', outfile);
  }
  fprintf(outfile, "#declare %s =\n", name);
  fprintf(outfile, "  union{\n");
  intersection = (intangent.x * (a2.x - a1.x) + intangent.y * (a2.y - a1.y) +
		  intangent.z * (a2.z - a1.z) < 0);
  if (tangentmode)
    OutputTorus(a2, a3, centre, radius, xang, yang, intangent, intersection);
  else
    OutputTorus(a1, a3, centre, radius, xang, yang, intangent, intersection);
  if (showtangents)
    OutputTangent(a1, intangent, a3, outtangent);
  a1 = a3;
  intangent = outtangent;
  scalevector(&intangent, -1.0);
  if (eoln(infile) & (feof(infile) == 0)) {
    fscanf(infile, "%*[^\n]");
	feof(infile);
    getc(infile);
  }
  while (feof(infile) == 0) {
    if (readcoords(&infile, &a2.x, &a2.y, &a2.z, s)) {
      addtopointlist(a2);
      if (eoln(infile) && (feof(infile) == 0)) {
        writevector("Control point: ", a2);
	fscanf(infile, "%*[^\n]");
	feof(infile);
	getc(infile);
      }
      else
	break;
      FindTorusCentre2(a1, a2, intangent, &centre, &outtangent, &radius,
		       &xang, &yang);
      intersection = (intangent.x * (a2.x - a1.x) + intangent.y * (a2.y - a1.y) +
		      intangent.z * (a2.z - a1.z) < 0);
      OutputTorus(a1, a2, centre, radius, xang, yang, intangent, intersection);
      if (showtangents)
	OutputTangent(a1, intangent, a2, outtangent);
      a1 = a2;
      intangent = outtangent;
      scalevector(&intangent, -1.0);
      continue;
    } else {
      if (*s != '\0' && toupper(s[0]) == 'R') {
	fscanf(infile, "%ld%ld%lg", &num_random, &randseed, &max_radius);
	       feof(infile);
	fscanf(infile, "%*[^\n]");
	       feof(infile);
	getc(infile);
	fscanf(infile, "%lg%lg%lg", &min.x, &min.y, &min.z);
	       feof(infile);
	fscanf(infile, "%*[^\n]");
	       feof(infile);
	getc(infile);
	fscanf(infile, "%lg%lg%lg", &max.x, &max.y, &max.z);
	       feof(infile);
	fscanf(infile, "%*[^\n]");
	       feof(infile);
	getc(infile);
	do {
	  num_random--;
	  repeats = 0;
	  do {
	    repeats++;
	    if (repeats % 10 == 0)
	      putchar('.');
	    GetRandomPoint(&a2, min, max, a1, 2 * max_radius);
	    FindTorusCentre2(a1, a2, intangent, &centre, &outtangent, &radius,
			     &xang, &yang);
	  } while (radius >= max_radius && repeats <= 500);
	  if (repeats > 500) {
	    putchar('\n');
	    printf("WARNING: Unable to meet maximum radius specification.\n");
	    printf("         Radius used: %12.6f\n", radius);
	  }
	  addtopointlist(a2);
	  writevector("Random control point: ", a2);
	  intersection = (intangent.x * (a2.x - a1.x) + intangent.y *
			    (a2.y - a1.y) + intangent.z * (a2.z - a1.z) < 0);
	  OutputTorus(a1, a2, centre, radius, xang, yang, intangent,
		      intersection);
	  if (showtangents)
	    OutputTangent(a1, intangent, a2, outtangent);
	  a1 = a2;
	  intangent = outtangent;
	  scalevector(&intangent, -1.0);
	} while (num_random != 0);
	continue;
      } else {
	while (*s != '\0' && (s[0] == ' ' || s[0] == ','))
	  strcpy(&s[0], &s[1]);
	if (*s != '\0')
	  continue;
	printf("ERROR: unable to parse input line.\n");
	if (infile != NULL)
	  fclose(infile);
	infile = NULL;
	if (outfile != NULL)
	  fclose(outfile);
	outfile = NULL;
	exit(0);
      }
    }
  }
  fprintf(outfile, "  }\n");
  putc('\n', outfile);
  if (showpoints) {
    fprintf(outfile, "#declare Points =\n");
    fprintf(outfile, "  union{\n");
    plist = &pointlisthead;
    do {
      WITH = plist;
      fprintf(outfile,
		     "    sphere{ <%12.6f, %12.6f, %12.6f> Point_Radius\n",
		     WITH->pt.x, WITH->pt.y,
		     WITH->pt.z);
      fprintf(outfile, "      texture{Point_Texture}\n");
      fprintf(outfile, "    }\n");
      plist = plist->next;
    } while (plist->next != NULL);
    fprintf(outfile, "  }\n");
  }
  if (outfile != NULL)
    fclose(outfile);
  outfile = NULL;
  if (infile != NULL)
    fclose(infile);
  infile = NULL;
  putchar('\n');
  printf("Patch file written successfully.\n");
  printf("Thanks for using TorPatch!\n");
  if (infile != NULL)
    fclose(infile);
  if (outfile != NULL)
    fclose(outfile);
  exit(1);
}
